import tensorflow.keras.backend as K
import tensorflow as tf
import numpy as np
from skimage.filters import gaussian
from skimage.morphology import label
from skimage.segmentation import find_boundaries
from matplotlib import cm


_jet = cm.get_cmap('jet')
_reds = cm.get_cmap('Reds')
_greens = cm.get_cmap('Greens')


def normalized_and_sort(val):
    mag = val.max(axis=1)
    idx = np.argsort(mag)[::-1]
    val = val / mag[:, None]
    val = val[idx]
    return val, mag


def summary_stat(val, stage, step=0):
    val, mag = normalized_and_sort(val)
    spc = val.mean(axis=1)
    tf.summary.histogram(f'max_val/{stage}', mag, step=step)
    tf.summary.histogram(f'avg_val/{stage}', spc, step=step)
    return val


def summary_spike(val, stage, step=0):
    tf.summary.image(f'spike/{stage}', _reds(val)[None, ...], step=step)


def summary_footprint_sample(val, mask, stage, step=0):
    h, w = mask.shape
    imgs = np.zeros((4, h, w))
    imgs[:, mask] = val[[0, 1, -2, -1]]
    tf.summary.image(
        f'footprint-sample/{stage}',
        _greens(imgs), max_outputs=4, step=step,
    )


def summary_footprint_max(val, mask, stage, step=0):
    h, w = mask.shape
    imgs_max = np.zeros((1, h, w))
    imgs_max[0, mask] = val.max(axis=0)
    tf.summary.image(f'max/{stage}', _greens(imgs_max), step=step)


def summary_segment(val, mask, flag, gauss, thr_out, stage):
    h, w = mask.shape
    out = 0
    b0 = False
    b1 = False
    for f, v in zip(flag[::-1], val[::-1]):
        img = np.zeros((h, w))
        img[mask] = v
        g = gaussian(img, gauss)
        g /= g.max()
        peak = np.argmax(g)
        y, x = peak // w, peak % w
        lbl = label(g > thr_out, connectivity=1)
        lbl = lbl == lbl[y, x]
        out += lbl
        if f:
            b0 |= find_boundaries(lbl)
        else:
            b1 |= find_boundaries(lbl)
    out = out / out.max()
    out = _greens(out)
    out[b0] = (0, 0, 0, 1)
    out[b1] = (1, 0, 0, 1)
    tf.summary.image(f'segment/{stage}', out[None, ...], step=0)



import tensorflow as tf
import numpy as np

from .base import Command, option, _option
from ..footprint.reduce import reduce_peak_idx
from ..footprint.make import make_segment
from ..train.summary import summary_segment
from ..util.tfrecord import load_tfrecord
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle


class SegmentCommand(Command):

    description = 'Make segment'

    name = 'segment'
    options = [
        _option('job-dir'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 2

    def is_target(self, stage):
        return stage == 2

    def force_stage(self, stage):
        return 2

    def create(self, data, prev, curr, logs, thr_dist):
        tfrecord = load_tfrecord(f'{data}-data')
        mask = load_numpy(f'{data}-mask')
        gauss, radius, shard = load_pickle(f'{prev}-filter')
        thr_out = self.status.params['thr-out']
        pos = load_numpy(f'{prev}-peak')
        score = load_numpy(f'{prev}-intensity')
        batch = self.status.params['batch']
        radius = np.array(radius)
        nr = radius.size
        writer = tf.summary.create_file_writer(logs)
        with writer.as_default():
            n = 10
            for i in range(6, n + 1):
                thr = i * thr_dist / n
                idx = reduce_peak_idx(pos, thr)
                r = radius[pos[idx, 1]]
                s = score[idx]
                tf.summary.histogram(f'radius/{curr[-3:]}', r, step=i)
                tf.summary.histogram(f'intensity/{curr[-3:]}', s, step=i)
                writer.flush()

            pos = pos[idx]
            cond = (0 < pos[:, 1]) & (pos[:, 1] < nr - 1)
            pos = pos[cond]
            r = radius[pos[:, 1]]
            s = s[cond]
            tf.summary.histogram(f'radius/{curr[-3:]}', r, step=n+1)
            tf.summary.histogram(f'intensity/{curr[-3:]}', s, step=n+1)
            writer.flush()

            segment = make_segment(
                tfrecord, mask, gauss, radius, pos, shard, batch,
            )
            nk = segment.shape[0]
            fsum = segment.sum(axis=1)
            tf.summary.histogram(f'sum_val/{curr[-3:]}', fsum, step=0)
            cond = np.ones(nk, np.bool)
            summary_segment(segment, mask, cond, gauss, thr_out, curr[-3:])
            writer.flush()
        writer.close()
        save_numpy(f'{curr}-segment', segment)
        save_numpy(f'{curr}-peak', pos[:, [0, 2, 3]])
        save_numpy(f'{curr}-radius', r)
        save_numpy(f'{curr}-intensity', s)



import tensorflow as tf
import pandas as pd
import numpy as np


def save_csv(base, peak, name):
    peak_dict = {k: v for k, v in zip(name, peak)}
    with tf.io.gfile.GFile(f'{base}.csv', 'w') as fp:
        pd.DataFrame(peak_dict).to_csv(fp)


def load_csv(base, name, typ):
    with tf.io.gfile.GFile(f'{base}.csv', 'r') as fp:
        peak = pd.read_csv(fp)
        peak = tuple(np.array(peak[k], t) for k, t in zip(name, typ))
    return peak



import tensorflow as tf
import pandas as pd
import numpy as np

from .base import Command, option, _option
from ..footprint.find import find_peak
from ..util.dataset import unmasked
from ..util.tfrecord import load_tfrecord
from ..util.numpy import load_numpy, save_numpy
from ..util.pickle import load_pickle, save_pickle


class PeakCommand(Command):

    description = 'Find peaks'

    name = 'peak'
    options = [
        _option('job-dir'),
        option('force', 'f'),
    ]

    def is_error(self, stage):
        return stage < 1

    def is_target(self, stage):
        return stage == 1

    def force_stage(self, stage):
        return 1

    def create(self, data, prev, curr, logs, gauss, radius, thr_gl, shard):
        tfrecord = load_tfrecord(f'{data}-data')
        mask = load_numpy(f'{data}-mask')
        nt = load_pickle(f'{data}-stat')[1]
        batch = self.status.params['batch']
        pos, score = find_peak(
            tfrecord, mask, gauss, radius, thr_gl, shard, batch, nt,
        )
        save_pickle(f'{curr}-filter', (gauss, radius, shard))
        save_numpy(f'{curr}-peak', pos)
        save_numpy(f'{curr}-intensity', score)
